const getEdgeKey = ({ from, to }) => btoa([from.id, to.id]).replaceAll(/[^A-Za-z0-9]+/g, '');

export class Vertex {
  constructor(graph, type, id, data, options = {}) {
    this.graph = graph;
    this.type = type;
    this.id = id;
    this.data = data;
    this.options = options;
    this.edges = {
      from: [],
      to: [],
    };
  }

  getEdges(type, away) {
    return this.edges[away ? 'from' : 'to'].filter(
      (edge) => edge.type === type,
    );
  }

  setDisplayLabel(label) {
    if (this.options.hide) {
      return;
    }
    this.graph.flowchart?.log(`${this.id}[${label}]`);
  }
}

export class Edge {
  constructor(graph, type, from, to, weight, data, options = {}) {
    this.graph = graph;
    this.from = from;
    this.to = to;
    this.type = type;
    this.weight = weight;
    this.data = data;
    this.options = options;
  }

  getHtml() {
    let html = '<table>';
    for (const { type, weight } of this.graph.getEdges(null, this.from, this.to)) {
      html += `<tr><td>${type}</td><td>${weight}</td></tr>`;
    }
    html += '</table>';
    return html;
  }

  getFlowchartNode() {
    return `${getEdgeKey(this)}(${this.getHtml()})`;
  }

  displayEdgeNode() {
    if (this.options.hide) {
      return;
    }
    this.graph.flowchart?.log(this.getFlowchartNode());
  }

  displayEdge() {
    if (this.options.hide) {
      return;
    }
    this.graph.flowchart?.log(`${this.from.id} --- ${this.getFlowchartNode()} --> ${this.to.id}`);
    this.graph.flowchart?.log(`class ${getEdgeKey(this)} edge`);
  }
}

export class WDAG {
  constructor(scene) {
    this.scene = scene;
    this.vertices = new Map();
    this.edgeTypes = new Map();
    this.nextVertexId = 0;
    this.flowchart = scene?.flowchart;
  }

  withFlowchart() {
    this.scene?.withAdditionalFlowchart();
    this.flowchart = this.scene?.lastFlowchart();
    return this;
  }

  addVertex(type, id, data, label, options) {
    // Support simple case of auto-incremented numeric ids
    if (typeof id === 'object') {
      data = id;
      id = this.nextVertexId++;
    }
    if (this.vertices.has(id)) {
      throw new Error(`Vertex already exists with id: ${id}`);
    }
    const vertex = new Vertex(this, type, id, data, options);
    this.vertices.set(id, vertex);
    vertex.setDisplayLabel(label ?? id);
    return vertex;
  }

  getVertex(id) {
    return this.vertices.get(id);
  }

  getVertexData(id) {
    return this.getVertex(id)?.data;
  }

  getVerticesData() {
    return Array.from(this.vertices.values()).map(({ data }) => data);
  }

  getEdge(type, from, to) {
    from = from instanceof Vertex ? from : this.getVertex(from);
    to = to instanceof Vertex ? to : this.getVertex(to);
    if (!from || !to) {
      return undefined;
    }
    const edges = this.edgeTypes.get(type);
    const edgeKey = getEdgeKey({ from, to });
    return edges?.get(edgeKey);
  }

  getEdgeWeight(type, from, to) {
    return this.getEdge(type, from, to)?.weight;
  }

  setEdgeWeight(type, from, to, weight, data, options) {
    from = from instanceof Vertex ? from : this.getVertex(from);
    to = to instanceof Vertex ? to : this.getVertex(to);
    const edge = new Edge(this, type, from, to, weight, data, options);
    let edges = this.edgeTypes.get(type);
    if (!edges) {
      edges = new Map();
      this.edgeTypes.set(type, edges);
    }
    const edgeKey = getEdgeKey(edge);
    edges.set(edgeKey, edge);
    edge.displayEdgeNode();
    return edge;
  }

  addEdge(type, from, to, weight, data, options) {
    from = from instanceof Vertex ? from : this.getVertex(from);
    to = to instanceof Vertex ? to : this.getVertex(to);
    if (this.getEdge(type, from, to)) {
      throw new Error(`Edge ${type} from ${from.id} to ${to.id} already exists`);
    }
    const edge = this.setEdgeWeight(type, from, to, weight, data, options);
    from.edges.from.push(edge);
    to.edges.to.push(edge);
    edge.displayEdge();
    return edge;
  }

  getEdges(type, from, to) {
    from = from instanceof Vertex ? from : this.getVertex(from);
    to = to instanceof Vertex ? to : this.getVertex(to);
    const edgeTypes = type ? [type] : Array.from(this.edgeTypes.keys());
    return edgeTypes.flatMap((edgeType) => {
      const edges = this.edgeTypes.get(edgeType);
      return Array.from(edges?.values() || []).filter((edge) => {
        const matchFrom = from === null || from === undefined || from === edge.from;
        const matchTo = to === null || to === undefined || to === edge.to;
        return matchFrom && matchTo;
      });
    });
  }

  countVertices(type) {
    if (!type) {
      return this.vertices.size;
    }
    return Array.from(this.vertices.values()).filter((vertex) => vertex.type === type).length;
  }
}
